#!/usr/bin/perl -w
# 
# OS Reviews CMS: content generator
# 
# Copyright (C) 2006, 2007, 2008 Hendrik Weimer, OS Reviews
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

use strict;
use DBI;
use XML::RSS;
use Image::Size;
use Net::NNTP;
use POSIX qw(strftime gmtime);

my ($dir, $file, $line, $document, %meta, @pro, @con, %knownmeta, $class, $i,
    %mode, @tmp, %template, $str, $j, $dbh, $ref, $latest, $k,
    $hottest, $rss, $total, $url, $caturl, $nntp, $nntpc, $maincat, %control, 
    $createtables, %config, %longcat, @wikipedia, $date, @compare, %tmpmeta);

my @possibleconfig = qw/\/etc\/osrcms.conf ~\/.osrcms.conf .\/osrcms.conf/;

my @reviewmeta = qw(name title author description category distro vgwort
		    homepage license version rating filename wikipedia);
my @topmeta = qw(filename title nolicenseboilerplate vgwort);
my @errormeta = qw(filename title);
my @miscmeta = qw(filename title author description category wikipedia vgwort);
my @comparisonmeta = qw(filename title author compare category wikipedia vgwort
                        description);

$longcat{'comparison'} = 'Comparison';

sub parse_config
{
    my ($name, $value);
    my $foundconfig = 0;
    my @allowednames = qw/artdir outdir templatedir searchdir commentdir
dbfile title shorttitle hostname itemsfront trackclasses commenthost nntphost
hierarchy controlemail controlname controlorg diggme slashdotme category logo
writearticles movefront itemsbest/;

    foreach $file (@possibleconfig)
    {
	if(open CONF, $file)
	{
	    $foundconfig = 1;
	    for($i=1; defined($line = <CONF>); $i++)
	    {
		# Get rid of comments and blank lines

		next if ($line =~ /^(\s*\#|\s*\n)/);

		$name = '';
		$value = '';
		
		$line =~ /^([A-Za-z]+)\s*=\s*'(.*)'\s*($|#)/;
		
		die "$file:$i: Invalid syntax!\n" if(!$1);
		die "$file:$i: Unknown variable: $1\n" 
		    if(!grep(/^$1$/, @allowednames));

		if($1 eq "category")
		{
		    if($2 =~ /(.*):(.*)/)
		    {
			$longcat{$1} = $2;
		    }
		    else
		    {
			die "$file:$i: Invalid category description\n";
		    }
		}
		else
		{
		    $config{$1} = $2;
		}
	    }
	}
    
    }

    die "No configuration file found!\n" if(!$foundconfig);

    # Default values

    $config{'movefront'} ||= 'index.html';
    $config{'diggme'} ||= 'false';
    $config{'slashdotme'} ||= 'false';
    $config{'itemsbest'} ||= 5;
    $config{'writearticles'} ||= 'true';

}
		
sub eigenlink
{
    my $str = $_[0];
    my $text = '';
    my ($name, $cat, $longname);
  
    $str =~ /(^|[^\\])\{(.*)\}/s;
    $name = $2;

    ($cat, $longname) = $dbh->selectrow_array("SELECT category, name FROM articles WHERE filename=" . $dbh->quote($name));

    if(!$cat)
    {
	print STDERR "$name: No such article\n";
	return $text;
    }
   
    if($str =~ /^\[(.*)\]/s)
    {
	$text = $1;
    }
    else
    {
	$text = $longname;
    }

    return "<a href=\"/reviews/$cat/$name\">$text</a>";
	
}

sub screenshot
{
    my ($big, $small, $desc) = @_;

    my ($x, $y, $err) = imgsize("$config{'outdir'}/images/articles/$small");

    die "$small: $err" if(!$x);

    $x .= 'px';
    
    my $str =  "<table class=\"screenshot\" style=\"width: $x\"><tr><td>";

    if($big)
    {
	return "$str<a href=\"/images/articles/$big\"><img src=\"/images/articles/$small\" alt=\"$desc\"></a></td></tr><tr><td><h3>$desc<br>(click to enlarge)</h3></td></tr></table><p>";
    }
    else
    {
	return "$str<img src=\"/images/articles/$small\" alt=\"$desc\"></td></tr><tr><td><h3>$desc</h3></td></tr></table><p>";
    }
}
    

sub makelist
{
    my $str = $_[0];
    my @items = split /\\item\s+/, $str;

    die "Missing \\item near $items[0]\n" if($items[0] =~ /\S/);

    shift @items;
    
    $str = join '</li><li>', @items;
    
    return '<li>' . $str . '</li>';
    
}

sub getbest
{
    my $mode = $_[0];
    my $ret = '';
    my $order = '';
    my $where;

    @tmp = split /,/, $config{'trackclasses'}; #/;

    foreach $i (@tmp)
    {
	$i = $dbh->quote($i);
    }
    
    $where = "class=" . join ' OR class=', @tmp;
	
    $order = "DESC"  if ($mode eq "published");

    $ref = $dbh->selectall_arrayref("SELECT name, filename, category, title, date(published) as date FROM articles WHERE $where AND $mode NOT NULL ORDER BY $mode $order LIMIT $config{'itemsbest'}", {'Slice' => {}});

    @tmp = @{$ref};

    for($i=0; $i<=$#tmp; $i++)
    {
	%meta = %{$tmp[$i]};
    
	$str = $template{'item_best'};
	$str =~ s/\$longcat/$longcat{$meta{'category'}}/g;
	$str =~ s/\$i/$i+1/ge;
	$str =~ s/\$hostname/$config{'hostname'}/g;
	$str =~ s/\$(\w+)/$meta{$1}/g
	    while($str =~ /\$/);
	
	$ret .= "$str";
    }

    return $ret;
}

sub newgroup
{
    my ($group, $tagline) = @_;
    my @control;
    my $date = `date -I`;
    chomp $date;

    push @control, "From: $config{'controlname'} <$config{'controlemail'}>\n";
    push @control, "Newsgroups: $config{'hierarchy'}.admin,$group\n";
    push @control, "Subject: cmsg newgroup $group\n";
    push @control, "Control: newgroup $group\n";
    push @control, "Message-ID: <newgroup-$group-$date-$$\@$config{'nntphost'}>\n";
    push @control, "Path: not-for-mail\n";
    push @control, "Date: " . `date --rfc-2822`;
    push @control, "Approved: $config{'controlemail'}\n";
    push @control, "Organization: $config{'controlorg'}\n";
    push @control, "\nFor your newsgroups file:\n$group\t$tagline\n";

    $nntp->post(\@control);
   
}

sub mathmode
{
    my $str = $_[0];

    $str =~ s/([A-Za-z]+)/<i>$1<\/i>/sg;
    $str =~ s/\\<i>([A-Za-z]+)<\/i>/\\$1/sg;
    $str =~ s/\^({[^\}+}]|.)/<sup>$1<\/sup>/sg;
    $str =~ s/(^|[^\\])\_({[^\}+}]|.)/<sub>$1<\/sub>/sg;
    $str =~ s/\\times/&times;/g;
    $str =~ s/\\(cos|sin|log|ln|exp)/$1/sg;
    
    return $str;
}

sub makelink
{
    my ($str, $url) = @_;

    # Remove math mode garbage from URL

    $url =~ s/<[A-Za-z\/]*>//g;
    
    return "<a href=\"$url\">$str</a>";

}

sub imgcat
{
    my $str = '';
    foreach $i (@tmp)
    {
	$str .= "<tr><td><a href=\"\/$j$i\/\"><img src=\"\/images\/$i.png\" alt=\"$longcat{$i}\" title =\"$longcat{$i}\"><\/a><\/td><\/tr>";
	
    }    
    return $str;
}

sub maketable
{
    my $str = $_[0];

    $str = "<table class=\"tabular\"><tr><td>$str";
    $str =~ s/\&/<\/td><td>/g;
    $str =~ s/\\\\/<\/td><\/tr><tr><td>/g;
    $str =~ s/\n+/\n/g;
    $str =~ s/<td>\s*\\multicolumn\{([^\}]*)\}\{([^\}]*)\}\{([^\}]*)\}/<td colspan="$1">$3/sg;
    $str .= '</td></tr>' if($str !~ /<\/td><\/tr>\s*$/);
    $str .= '</table>';
    
    return $str;
}

sub parse_file
{
    my ($file, $opt) = @_;

    open FILE, $file  or die "$file: $!";

    $mode{'document'} = 0;
    
    $i = 1;
    undef @pro;
    undef @con;
    undef %tmpmeta;
    if(!$opt)
    {
	undef @wikipedia;
	undef @compare;
	$document = '';
    }
    
    while(defined($line = <FILE>))
    {
	# Strip comments 
	
	$line =~ s/(^|[^\\])%.*$/$1/g;
	
	if(!$mode{'document'})
	{
	    if($line =~ /\{/)
	    {
		$line .= <FILE> while($line !~ /\}/);
	    }
	    
	    if ($line =~ /\\begin\{document\}/)
	    {
		$mode{'document'} = 1;
	    }
	    
	    elsif($line =~ /\\documentclass\{(.*?)\}/)
	    {
		$tmpmeta{'class'} = $1;
		$class = $1 if(!$opt);
	    }
	    
	    elsif($line =~ /\\pro\{(.*?)\}/s)
	    {
		push @pro, $1;
	    }

	    elsif($line =~ /\\con\{(.*?)\}/s)
	    {
		push @con, $1;
	    }
	    
	    elsif($line =~ /\\wikipedia\{(.*?)\}/s)
	    {
		push @wikipedia, $1 if(!$opt);
	    }
	    
	    elsif($line =~ /\\compare\{(.*?)\}/s)
	    {
		push @compare, $1 if(!$opt);
	    }
	    
	    elsif ($line =~ /\\(.*?)\{(.*?)\}/s)
	    {
		die("$file:$i: command $1 not known for class $class.\n")
		    if(!grep(/^$1$/, @{$knownmeta{$tmpmeta{'class'}}}));
		
		$tmpmeta{$1} = $2;
		
	    }
	    elsif ($line =~ /\\(.*)/)
	    {
		die("$file:$i: command $1 not known for class $class.\n")
		    if(!grep(/^$1$/, @{$knownmeta{$tmpmeta{'class'}}}));
		$tmpmeta{$1} = 1;
	    }

	}
	else
	{
	    if ($line =~ /\\end\{document\}/)
	    {
		$mode{'document'} = 0;
	    }
	    else
	    {
		$document .= $line if (!$opt);
	    }
	    
	}
	
	$i++;
    }
    
    close FILE;

    %meta = %tmpmeta if(!$opt);
}

sub make_reviewbox
{
    my %meta = %{$_[0]};
    my $opt = $_[1];

    $meta{'stable'} = $meta{'unstable'} = $meta{'fedora'}
    = $meta{'mandriva'} = $meta{'suse'} = $meta{'ubuntu'} 
    = "&#x25A1;";
	    
    my $str = $template{'rbox'};

    $str = $template{'cbox_item'} if($opt);

    @tmp = split /,/, $meta{'distro'}; #/;
	    
    foreach $i (@tmp)
    {
	$meta{$i} = "&#x25A0;";
    }
    
    $meta{'pros'} = join '</li><li>', @pro;
    $meta{'cons'} = join '</li><li>', @con;
    
    $meta{'rating'} = "0$meta{'rating'}" if $meta{'rating'} < 10;
    
    $str =~ s/(^|[^\\])\$([A-Za-z0-9_]+)/$1$meta{$2}/g;

    return $str;

}

&parse_config;

$knownmeta{'review'} = \@reviewmeta;
$knownmeta{'misc'} = \@miscmeta;
$knownmeta{'error'} = \@errormeta;
$knownmeta{'top'} = \@topmeta;
$knownmeta{'comparison'} = \@comparisonmeta;

if($config{'commenthost'})
{
    $nntp = Net::NNTP->new($config{'nntphost'});
    $nntp->reader;
}

opendir(DIR, "$config{'templatedir'}") or die "$config{'templatedir'}: $!\n";

while(defined($file = readdir(DIR)))
{
    next if $file !~ /^(.*)\.html$/;

    open FILE, "$config{'templatedir'}/$file" or 
	die "$config{'templatedir'}/$file: $!\n";

    while(defined($line = <FILE>))
    {
	$template{$1} .= $line;
    }
    close FILE;
}

closedir(DIR);

foreach $i (keys %template)
{
    $j = 0;
    while($template{$i} =~ /\$template\{(\w+)\}/)
    {
	$j++;
	$template{$i} =~ s/\$template\{(\w+)\}/$template{$1}/g;
	die "Recursion in template $1\n" if $j > 10;
    }
	
}

`rm -rf $config{'outdir'}`;
mkdir "$config{'outdir'}" or die "$!\n";


foreach $i qw(reviews css images images/rating images/articles)
{
    if(! -d "$config{'outdir'}/$i")
    {
	mkdir "$config{'outdir'}/$i" or die "$config{'outdir'}/$i: $!\n";
	`cp $config{'templatedir'}/htaccess $config{'outdir'}/$i/.htaccess`;
    }
}

`cp $config{'templatedir'}/htaccess.top $config{'outdir'}/.htaccess`;

`cp css/*.css $config{'outdir'}/css`;
`cp images/*.png images/favicon.ico $config{'outdir'}/images`;
`cp images/rating/*.png $config{'outdir'}/images/rating`;

$dbh = DBI->connect("DBI:SQLite:$config{'dbfile'}") or die;

{
    local $dbh->{PrintError} = 0;
    local $dbh->{RaiseError} = 1;
    eval {$dbh->selectrow_array("SELECT * FROM articles LIMIT 1")};
    if($@)
    {
	$createtables = 1;
    }
}

if($createtables)
{
    $dbh->do('BEGIN');
    $dbh->do('CREATE TABLE articles (	
	id		INTEGER UNIQUE PRIMARY KEY,
	name		VARCHAR(40),
	filename        VARCHAR(20),
	title		VARCHAR(80),
	description     VARCHAR(1024),
	category        VARCHAR(20),
	class           VARCHAR(20),
	published	TIMESTAMP,
	svn		INT,
	rank		INT
	)');
    $dbh->do('CREATE INDEX art1 ON articles (published)');
    $dbh->do('CREATE INDEX art2 ON articles (filename)');
    $dbh->do('CREATE INDEX art3 ON articles (class)');
    $dbh->do('CREATE INDEX art4 ON articles (category)');

    $dbh->do('CREATE TABLE categories (
	id		INTEGER UNIQUE PRIMARY KEY,
	filename	VARCHAR(20),
	category	VARCHAR(20)
	)');
    $dbh->do('CREATE INDEX cat1 ON categories (filename)');
    $dbh->do('CREATE INDEX cat2 ON categories (category)');

    $dbh->do('COMMIT');
}

$latest = getbest("published");
$hottest = getbest("rank");

opendir(DIR, $config{'artdir'}) or die "$config{'artdir'}: $!\n";

while(defined($dir = readdir(DIR)))
{
    next if $dir =~ /^\./;

    `cp $config{'artdir'}/$dir/*.png $config{'outdir'}/images/articles 2>/dev/null`;
    `cp $config{'artdir'}/$dir/*.jpg $config{'outdir'}/images/articles 2>/dev/null`;

    opendir(SUBDIR, "$config{'artdir'}/$dir") or die "$config{'artdir'}/$dir: $!\n";
    
    while(defined($file = readdir(SUBDIR)))
    {
	next if $file !~ /\.tex$/;
	
	parse_file "$config{'artdir'}/$dir/$file";

	$document = "<p>$document</p>";

	# Links (first pass)

	$document =~ s/\\link\{([^\}]*)\}\{([^\}]*)\}/makelink($1, $2)/seg;

	$document =~ s/\\eigenlink(.*?\})/eigenlink($1)/sge;

	# Math mode

	$document =~ s/([^\\])\$(.*?)([^\\])\$/$1.mathmode($2.$3)/se
	while($document =~ /[^\\]\$/);
	$document =~ s/\\\$/\$/g;

	# Commands
	
	$document =~ s/\\latest(\W)/$latest$1/g;

	# Tables

	$document =~
	    s/\\begin\{tabular\}\{(.*?)\}(.*?)\\end\{tabular\}/maketable($2, $1)/se while($document =~ /\\begin\{tabular\}\{.*?\}(.*?)\\end\{tabular\}/s);

	# Physical text attributes

	$document =~ s/\{\s*\\tt( |\n)(.*?[^\\])\}/<tt>$2<\/tt>/sg;
	$document =~ s/\{\s*\\it( |\n)(.*?)\}/<i>$2<\/i>/sg;
	$document =~ s/\{\s*\\bf( |\n)(.*?)\}/<b>$2<\/b>/sg;

	$document =~ s/\`\`|\'\'/&quot;/g;
	$document =~ s/[^\\]\~|\'\'/&nbsp;/g;
	$document =~ s/([^\\])\\ /$1&nbsp;/g;
	$document =~ s/\\\~/~/g;
	$document =~ s/(\s)--(\s)/$1&mdash;$2/g;
	$document =~ s/\\\\(\s|$)/<br>/g;
	$document =~ s/\n\s*\n/<\/p><p>/sg;

	# Special chars

	$document =~ s/\\'(.)/&$1acute;/g;
	$document =~ s/\\"(.)/&$1uml;/g;
	$document =~ s/\\&/&/g;
	$document =~ s/\\_/_/g;
	$document =~ s/\\#/#/g;
	$document =~ s/\\%/%/g;
	$document =~ s/\\{/{/g;
	$document =~ s/\\}/}/g;
	
	# Environments

	$document =~ s/\\begin\{itemize\}(.*?)\\end\{itemize\}/'<\/p><ul>' . makelist($1). '<\/ul><p>'/se
	    while($document =~ /\\begin\{itemize\}(.*?)\\end\{itemize\}/s);

	$document =~ s/\\begin\{enumerate\}(.*?)\\end\{enumerate\}/'<\/p><ol>' . makelist($1). '<\/ol><p>'/se
	    while($document =~ /\\begin\{enumerate\}(.*?)\\end\{enumerate\}/s);

	# Links (second pass)

	$document =~ s/\\link\{([^\}]*)\}\{([^\}]*)\}/makelink($1, $2)/seg;

	$document =~ s/\\eigenlink(.*?\})/eigenlink($1)/sge;

	# Screenshots

	$document =~ s/<p>\\screenshot{(.*?)}{(.*?)}{(.*?)}/screenshot($1, $2, $3)/sge;

	# Sections

	$document =~ s/<p>(\n)?\\chapter{(.*?)}<\/p>/<h2>$2<\/h2>/sg;
	$document =~ s/<p>(\n)?\\section{(.*?)}<\/p>/<h3>$2<\/h3>/sg;
	$document =~ s/<p>(\n)?\\subsection{(.*?)}<\/p>/<h4>$2<\/h4>/sg;

	if($document =~ /^(.*)\\readmore/s)
	{
	    $meta{'description'} = "$1 <span class=\"readmore\"><a href=\"\$url\">Read more</a></span>";
	    $document =~ s/\\readmore//g;
	}

	# Remove clutter

	$document =~ s/<p><table( |>)/<table$1/g;
	$document =~ s/<\/table><\/p>/<\/table>/g;

	if ($class eq "review")
	{
	    $maincat = (split /,/, $meta{'category'})[0]; # /);
			
	    $document = "<!-- PRINT -->$document";

	    # Link to comment system

	    $document .= "<p><b><a href=\"http://$config{'commenthost'}/comments?article=reviews.$maincat.$meta{'filename'}\">View comments</a></b></p>" if($config{'commenthost'});
	
	    # Review box
	    
	    $document .= make_reviewbox(\%meta)
	}
	
	elsif($class eq "misc")
	{
	    $maincat = (split /,/, $meta{'category'})[0]; # /);

	    $document = "<!-- PRINT -->$document";

	    $document .= "<p><b><a href=\"http://$config{'commenthost'}/comments?article=$maincat.$meta{'filename'}\">View comments</a></b></p>" if($config{'commenthost'});
	}

	elsif($class eq "comparison")
	{
	    $maincat = (split /,/, $meta{'category'})[0]; # /);

	    $document = "<!-- PRINT -->$document";

	    $document .= "<p><b><a href=\"http://$config{'commenthost'}/comments?article=reviews.comparison.$meta{'filename'}\">View comments</a></b></p>" if($config{'commenthost'});
	    
	    $str = $template{'cbox'};

	    foreach $file (@compare)
	    {
		parse_file "$config{'artdir'}/$file", 1;

		$tmpmeta{'homepage'} =~ s/http:\/\///g;
		$tmpmeta{'homepage'} =~ s/\/$//g 
		    if($tmpmeta{'homepage'} !~ /\/./);
		    
		$tmpmeta{'stable'} = $tmpmeta{'unstable'} = $tmpmeta{'fedora'}
		= $tmpmeta{'mandriva'} = $tmpmeta{'suse'} = $tmpmeta{'ubuntu'} 
		= '<div class="distrobox">&#x25A1;</div>';
		
		@tmp = split /,/, $tmpmeta{'distro'}; #/;
	    
		foreach $i (@tmp)
		{
		    $tmpmeta{$i} = '<div class="distrobox">&#x25A0;</div>';
		}

		$tmpmeta{'pros'} = '<ul><li>'. join('</li><li>', @pro) 
		    . '</li></ul>';
		$tmpmeta{'cons'} = '<ul><li>'. join('</li><li>', @con) 
		    . '</li></ul>';
    
		$tmpmeta{'rating'} = "0$tmpmeta{'rating'}" if $tmpmeta{'rating'} < 10;

		$str =~ s/\$name/<th class="name">$tmpmeta{'name'}<\/th>\$name/g;
		$str =~ s/\$version/<th class="version">$tmpmeta{'version'}<\/th>\$version/g;
		$str =~ s/\$homepage/<td><a href="http:\/\/$tmpmeta{'homepage'}">$tmpmeta{'homepage'}<\/a><\/td>\$homepage/g;
		$str =~ s/\$(license|stable|unstable|fedora|mandriva|suse|ubuntu|pros|cons)/<td>$tmpmeta{$1}<\/td>\$$1/g;
		
		$str =~ s/\$rating/<td><div class="crating"><img src="\/images\/rating\/$tmpmeta{'rating'}.png" width="72" height="72" alt="$tmpmeta{'rating'}"><\/div><\/td>\$rating/g;
		
	    }

	    $str =~ s/\$\w+//g;

	    $document .= $str;

	    $meta{'name'} = $longcat{'comparison'};
	}

	# VG Wort counter
	
	$document .= '<img src="http://vg00.met.vgwort.de/na/' 
	    . $meta{'vgwort'} . '" width="1" height="1" alt="">'
	    if($meta{'vgwort'});
       
	# Document management

	if($class eq "review" or $class eq "misc" or $class eq "comparison")
	{
	    $j = '';
	    $j = "reviews/" if ($class eq "review" or $class eq "comparison");

	    $meta{'name'} = $longcat{$maincat} if($class eq "misc");

	    $date = $dbh->selectrow_array("SELECT date(published) FROM articles where filename=" . $dbh->quote($meta{'filename'}));
	    $date ||= `date -I`;

	    @tmp = split /,/, $meta{'category'}; # /;
	    unshift @tmp, 'comparison' if($class eq 'comparison');

	    $maincat = $tmp[0];
	    
	    $document .= "<p class=\"license\">$template{'license'}</p>";

	    $str = $template{'print'};
	    $str ||= '';

	    $str =~ s/\$longcat/$longcat{$maincat}/g;
	    $str =~ s/\$date/$date/g;
	    $str =~ s/\$(\w+)/$meta{$1}/g
		while($str =~ /\$/);

	    $str =~ s/\+\+\+CONTENT\+\+\+/$document/;
	    $str =~ s/\+\+\+TITLE\+\+\+/$config{'title'}/;
	    $str =~ s/<!-- PRINT -->/<p class="date"><a href="\/$j$maincat\/$meta{'filename'}">Normal version<\/a><\/p>/;
	    
	    $file = "$config{'outdir'}/$j$maincat/" . $meta{'filename'} . '.print';

	    foreach $k (@tmp)
	    {
		if (! -d "$config{'outdir'}/$j$k")
		{
		    mkdir "$config{'outdir'}/$j$k" or die "$!\n";
		    `cp $config{'templatedir'}/htaccess $config{'outdir'}/$j$k/.htaccess`;
		}
	    }

	    if($config{'writearticles'} ne 'false')
	    {
		open FILE, ">$file" or die "$file: $!\n";
	    
		print FILE $str;
	    
		close FILE;
	    }

	    $str = $template{$class};
	    $str ||= '';

	    $str =~ s/\@catimages/'<div class="category"><table>' . &imgcat  . '<\/table><\/div>'/sge; # /;
	
    
	    $str =~ s/\$longcat/$longcat{$maincat}/g;
	    $str =~ s/\$date/$date/g;
	    $str =~ s/\$(\w+)/$meta{$1}/g
		while($str =~ /\$/);
	    
		
	    $str =~ s/\+\+\+CONTENT\+\+\+/$document/;
	    $str =~ s/\+\+\+LATEST\+\+\+/$latest/;
	    $str =~ s/\+\+\+HOTTEST\+\+\+/$hottest/;
            $str =~ s/\+\+\+TITLE\+\+\+/$config{'shorttitle'} - $meta{'name'}: $meta{'title'}/;
	    
	    if(@wikipedia)
	    {
		foreach $i (@wikipedia)
		{
		    $k = $i;
		    $k =~ s/ /_/g;
		    $i = "<a href=\"http://en.wikipedia.org/wiki/$k\">$i</a>";
		}
		$str =~ s/\+\+\+WIKIPEDIA\+\+\+/'<ul><li><h2>Wikipedia<\/h2><\/li><li><table><tr><td>' . join('<\/td><\/tr><tr><td>', @wikipedia) . '<\/td><\/tr><\/table><\/li><\/ul>'/ge;

	    }
	    else
	    {
		$str =~ s/\+\+\+WIKIPEDIA\+\+\+//g;
	    }
	    
	    $i = "<p class=\"date\"><a href=\"/$j$maincat\/$meta{'filename'}.print\">Print version</a>";
	    $i .= " - <a href=\"http://www.digg.com/submit?phase=2&amp;url=http://$config{'hostname'}/$j$maincat/$meta{'filename'}&amp;ei=utf-8\">Digg this</a>"
		if($config{'diggme'} eq "true");
	    $i .= " - <a href=\"http://slashdot.org/submit.pl\">Slashdot this</a>" 
		if($config{'slashdotme'} eq "true");
	    
	    $str =~ s/<!-- PRINT -->/$i<\/p>/;

	    $file = "$config{'outdir'}/$j$maincat/" . $meta{'filename'};

	    if($config{'writearticles'} ne 'false')
	    {
		open FILE, ">$file" or die "$file: $!\n";
	    
		print FILE $str;
	    
		close FILE;
	    }
	    
	    $j =~ s/\//./;
	    
	    newgroup("$config{'hierarchy'}.$j$maincat.$meta{'filename'}", 
		     "$meta{'name'}: $meta{'title'}.")
		if ($config{'commenthost'} 
		    && !$nntp->group("$config{'hierarchy'}.$j$tmp[0]" 
				     . ".$meta{'filename'}"));
	
	    foreach $i (keys %meta)
	    {
		$meta{$i} = $dbh->quote($meta{$i});
	    }

	    $maincat = $dbh->quote($maincat);
	    
	    if(!$dbh->selectrow_array("SELECT name FROM articles where filename= $meta{'filename'}"))
	    {
		$dbh->do("INSERT INTO articles (name, filename, title, description, category, published, class) VALUES ($meta{'name'}, $meta{'filename'}, $meta{'title'}, $meta{'description'}, $maincat, datetime('now'), $meta{'class'})\n");

	    }
	    else
	    {
		$dbh->do("UPDATE articles SET name = $meta{'name'}, title = $meta{'title'}, description = $meta{'description'}, category = $maincat, class = $meta{'class'} WHERE filename = $meta{'filename'}");
	    }

	    $dbh->do("DELETE FROM categories WHERE filename=$meta{'filename'}");
	    
	    for($i=0; $i<=$#tmp; $i++) 
	    {
		$tmp[$i] = $dbh->quote($tmp[$i]);
		$dbh->do("INSERT INTO categories (filename, category) VALUES ($meta{'filename'}, $tmp[$i])");
	    }
		   
	    
	}

        elsif($class eq "top")
        {
	    $document = "<h2>$meta{'title'}</h2><p></p>$document";
	    $document .= "<p class=\"license\">$template{'license'}</p>" 
		if(!$meta{'nolicenseboilerplate'});

            $str = $template{'top'};

	    $str =~ s/\$date/$date/g;
	    $str =~ s/\$(\w+)/$meta{$1}/g
		while($str =~ /\$/);

            $str =~ s/\+\+\+CONTENT\+\+\+/$document/;
            $str =~ s/\+\+\+LATEST\+\+\+/$latest/;
            $str =~ s/\+\+\+HOTTEST\+\+\+/$hottest/;
	    $str =~ s/\+\+\+TITLE\+\+\+/$config{'title'}/;

	    $file = "$config{'outdir'}/$meta{'filename'}";

	    open FILE, ">$file" or die "$file: $!\n";
    
	    print FILE $str;
	    
	    close FILE;
	}

        elsif($class eq "error")
        {
	    $document = "<h2>$meta{'title'}</h2><p></p>$document";

            $str = $template{'error'};

            $str =~ s/\+\+\+CONTENT\+\+\+/$document/;
            $str =~ s/\+\+\+LATEST\+\+\+/$latest/;
            $str =~ s/\+\+\+HOTTEST\+\+\+/$hottest/;
	    $str =~ s/\+\+\+TITLE\+\+\+/$meta{'title'}/;

	    $file = "$config{'outdir'}/$meta{'filename'}";

	    open FILE, ">$file" or die "$file: $!\n";
    
	    print FILE $str;
	    
	    close FILE;
	}

    }
    
    closedir SUBDIR;
}

closedir DIR;

for $j (@{$dbh->selectcol_arrayref("SELECT category from categories GROUP BY category")})
{
    $j = $dbh->quote($j);
    $document = '';

    $ref = $dbh->selectall_arrayref("SELECT *, strftime(\"%s\", published) as pubdate, strftime(\"%Y-%m-%d %H:%M\", published) as pubdate2, articles.category as maincat FROM articles, categories WHERE articles.filename=categories.filename AND categories.category=$j ORDER by articles.name, articles.title DESC",
				    {'Slice' => {}});

    for($i=0; $i<=$#$ref; $i++)
    {
	%meta = %{(@{$ref})[$i]};

	if($meta{'class'} eq "misc")
	{
	    $meta{'url'} = "/$meta{'maincat'}/$meta{'filename'}";
	    $meta{'caturl'} = "/$meta{'maincat'}/";
	}
	else
	{
	    $meta{'url'} = "/reviews/$meta{'maincat'}/$meta{'filename'}";
	    $meta{'caturl'} = "/reviews/$meta{'maincat'}/";
	}

	    
	$str = $template{"item_$meta{'class'}"};
	$str ||= '';
	$str =~ s/\$longcat/$longcat{$meta{'category'}}/g;
	$str =~ s/\$(\w+)/$meta{$1}/g
	    while($str =~ /\$/);

	$document .= "<li>$str</li>";

    }
    $document =~ s/\`\`|\'\'/&quot;/g;
    $document .= "<li class=\"license\">$template{'license'}</li>";
	
    $str = $template{'index'};
	
    $str =~ s/\+\+\+CONTENT\+\+\+/$document/;
    $str =~ s/\+\+\+LATEST\+\+\+/$latest/;
    $str =~ s/\+\+\+HOTTEST\+\+\+/$hottest/;

    $meta{'category'} = "reviews/$meta{'category'}" 
	if($meta{'class'} eq 'review' or $meta{'class'} eq 'comparison');    
    $file = "$config{'outdir'}/$meta{'category'}/index.html";

    if($config{'writearticles'} ne 'false')
    {
	open FILE, ">$file" or die "$file: $!\n";
	    
	print FILE $str;

	close FILE;
    }
}
    


$total = $dbh->selectrow_array("SELECT COUNT(*) FROM articles");

for($j=0; $j<$total; $j+=$config{'itemsfront'})
{
    $ref = $dbh->selectall_arrayref('SELECT name, filename, title, category, description, strftime("%s", published) as pubdate, strftime("%Y-%m-%d %H:%M", published) as pubdate2, date(published) as date, class FROM articles ORDER BY published DESC LIMIT ' . "$config{'itemsfront'} OFFSET $j", {'Slice' => {}});

    $document = '';

    if(!$j)
    {
	@tmp = @{$ref};
	$str = strftime('%a, %d %b %Y %H:%M:%S UT', 
			gmtime(${%{$tmp[0]}}{'pubdate'}));
	
	$rss = new XML::RSS (version => '2.0');

	$rss->channel(title => "$config{'shorttitle'}",
		      link =>  "http://$config{'hostname'}",
		      language => 'en-US',
		      description => "$config{'title'}",
		      lastBuildDate => $str,
		      ttl => 120);

	$config{'logo'} ||= '/images/logo3.png';
	
	$rss->image(title => "$config{'shorttitle'}",
		    url => "http://$config{'hostname'}$config{'logo'}",
		    link => "http://$config{'hostname'}" ,
		    description => "$config{'shorttitle'}");
    }
    
    foreach $i (@{$ref})
    {
	%meta = %{$i};
	$meta{'pubdate'} = strftime('%a, %d %b %Y %H:%M:%S UT', 
				    gmtime($meta{'pubdate'}));

	if($j + $config{'itemsfront'} > $total && ${$ref}[$#$ref]{name} eq $meta{name})
	{
	    $document .= "<li style=\"border-bottom: 0;\">";
	}

	else
	{
	    $document .= "<li>";
	}

	$str = $template{'item_front'};

	if($meta{'class'} eq "misc")
	{
	    $meta{'url'} = "/$meta{'category'}/$meta{'filename'}";
	    $meta{'caturl'} = "/$meta{'category'}/";
	    $str =~ s/\$longcat/$longcat{$meta{'category'}}/g;

	}
	else
	{
	    $meta{'url'} = "/reviews/$meta{'category'}/$meta{'filename'}";
	    $meta{'caturl'} = "/reviews/$meta{'category'}/";
	    $str =~ s/\$longcat/$longcat{$meta{'category'}}/g;
	}

	$str =~ s/\$(\w+)/$meta{$1}/g
	    while($str =~ /\$/);

	if($str =~ /\@categories/)
	{
	    @tmp = @{$dbh->selectcol_arrayref("SELECT category FROM categories where filename=" . $dbh->quote($meta{'filename'}))};
	    foreach $k (@tmp)
	    {
		if($meta{'class'} eq "misc")
		{
		    $k = "<a href=\"/$k\">$k</a>";
		}
		else
		{
		    $k = "<a href=\"/reviews/$k\">$k</a>";
		}
	    }
	    $str =~ s/\@categories/join(', ', @tmp)/ge;
	}
   	
	$document .= $str . '</li>';

	if(!$j)
	{
	    $meta{'description'} =~ s/<[^>]*>//sg;
	    if($config{'writearticles'} eq 'false')
	    {
		$rss->add_item(title => "$meta{'title'}",
			       link => "http://" . $config{'hostname'} 
			       . "/$config{'movefront'}",
			       description => $meta{'description'},
			       pubDate => $meta{'pubdate'});
	    }
	    else
	    {
		$rss->add_item(title => "$meta{'name'}: $meta{'title'}",
			       link => "http://" . $config{'hostname'} 
			       . $meta{'url'},
			       description => $meta{'description'},
			       pubDate => $meta{'pubdate'});

	    }
	}
    }

    if(!$j)
    {
	$rss->save("$config{'outdir'}/rss.xml");
    }

    if($j + $config{'itemsfront'} < $total)
    {
        $document .= '<li class="final"><a href="/' 
	    . ($j + $config{'itemsfront'}) . '">Older articles</a></li>';
    }
    
    $document =~ s/\`\`|\'\'/&quot;/g;
    $document .= "<li class=\"license\">$template{'license'}</li>";
    
    $str = $template{'index'};
    
    $str =~ s/\+\+\+CONTENT\+\+\+/$document/;
    $str =~ s/\+\+\+LATEST\+\+\+/$latest/;
    $str =~ s/\+\+\+HOTTEST\+\+\+/$hottest/;

    if($j)
    {
        $file = "$config{'outdir'}/$j";
    }
    else
    {
        $file = "$config{'outdir'}/$config{'movefront'}";
    }

    open FILE, ">$file" or die "$file: $!\n";
    
    print FILE $str;
    
    close FILE;

}

# Search template

if($template{'estseek'})
{
    $str = $template{'estseek'};

    $str =~ s/\+\+\+LATEST\+\+\+/$latest/;
    $str =~ s/\+\+\+HOTTEST\+\+\+/$hottest/;
    
    $file = "$config{'searchdir'}/estseek.tmpl";
    
    open FILE, ">$file" or die "$file: $!\n";

    print FILE $str;
    
    close FILE;
}

# Comment system template

if($config{'commenthost'})
{
    $str = $template{'comments'};

    $str =~ s/\+\+\+LATEST\+\+\+/$latest/;
    $str =~ s/\+\+\+HOTTEST\+\+\+/$hottest/;

    $file = "$config{'commentdir'}/comments.html";

    open FILE, ">$file" or die "$file: $!\n";

    print FILE $str;

    close FILE;

    `cp comments comments.conf $config{'commentdir'}`;
}

$nntp->quit if($config{'commenthost'});

